from change_case import ChangeCase


def db_table_name(table_name, schema_name=None):
    table_name = ChangeCase.camel_to_snake(table_name)
    if schema_name is None:
        return f'"{table_name}"'
    else:
        return f'"{schema_name}"."{table_name}"'


def db_column_name(col_name):
    return ChangeCase.camel_to_snake(col_name)



from enum import Enum


# TODO: Define the types as enum
class Type(Enum):
    pass


POSTGRES_TYPES = {
    'boolean': 'bool',
    'number': 'float',
    'string': 'varchar({})',
    'enum': 'text',
    'integer': 'bigint',
    'timestamp': 'timestamptz',
    'date': 'date',
    'link': 'integer',
    'object': 'json',
    'id': 'serial',
}

REDSHIFT_TYPES = {
    **POSTGRES_TYPES,
    'object': 'text',
    'id': 'int identity(1, 1) not null',
}

COLUMNS_TYPES = {
    'postgres': POSTGRES_TYPES,
    'redshift': REDSHIFT_TYPES,
}

FK_TYPES = {
    'serial': 'bigint',
    'int identity(1, 1) not null': 'bigint',
}



import logging
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Set

from jsonschema2ddl.types import COLUMNS_TYPES, FK_TYPES
from jsonschema2ddl.utils import db_column_name


@dataclass
class Column:
    name: str
    database_flavor: str = "postgres"
    comment: str = field(default_factory=str, repr=False)
    constraints: Dict = field(default_factory=dict, repr=False)
    jsonschema_type: str = field(default_factory=str, repr=False)
    jsonschema_fields: Dict = field(default_factory=dict, repr=False)

    logger: logging.Logger = field(default=logging.getLogger('Column'), repr=False)

    @property
    def max_lenght(self) -> int:
        return self.jsonschema_fields.get('maxLength', 256)

    @property
    def data_type(self) -> str:
        if 'format' in self.jsonschema_fields:
            # FIXME: catch this case as a more generic type
            if self.jsonschema_fields['format'] == 'date-time':
                return 'timestamptz'
            elif self.jsonschema_fields['format'] == 'date':
                return 'date'
        return COLUMNS_TYPES[self.database_flavor][self.jsonschema_type].format(self.max_lenght)

    # FIXME: Property or simple function?
    @property
    def is_pk(self) -> bool:
        return self.jsonschema_fields.get('pk', False)

    # FIXME: Property or simple function?
    @property
    def is_index(self) -> bool:
        return self.jsonschema_fields.get('index', False)

    # FIXME: Property or simple function?
    @property
    def is_unique(self) -> bool:
        return self.jsonschema_fields.get('unique', False)

    @staticmethod
    def is_fk() -> bool:
        """Returns true if the column is a foreign key.

        Returns:
            bool: True if it is foreign key
        """
        return False

    def __hash__(self):
        return hash(self.name)

    # FIXME: Avoid overwritting the the repr method
    # NOTE: Overwrite dataclass method to show data_type property
    def __repr__(self):
        return f"Column(name={self.name} data_type={self.data_type})"


@dataclass
class Table:
    ref: str
    name: str
    database_flavor: str = "postgres"
    columns: Set[Column] = field(default_factory=set)
    primary_key: Column = None
    comment: str = None
    indexes: List[str] = field(default_factory=list)
    unique_columns: List[str] = field(default_factory=list)
    jsonschema_fields: Dict = field(default_factory=dict, repr=False)

    logger: logging.Logger = field(default=logging.getLogger('Table'), repr=False)
    _expanded: bool = field(default=False, repr=False)

    def expand_columns(
            self,
            table_definitions: Dict = dict(),
            columns_definitions: Dict = dict(),
            referenced: bool = False) -> Dict:
        if self._expanded:
            self.logger.info('Already expanded table. Skiping...')
            return self
        for col_name, col_object in self.jsonschema_fields.get('properties').items():
            self.logger.debug(f'Creating column {col_name}')
            col_name = db_column_name(col_name)
            self.logger.debug(f'Renamed column to {col_name}')
            if '$ref' in col_object:
                self.logger.debug(f"Expanding {col_name} reference {col_object['$ref']}")
                self.logger.debug(table_definitions)
                if col_object['$ref'] in table_definitions:
                    ref = col_object['$ref']
                    self.logger.debug(f'Column is a FK! Expanding {ref} before continue...')
                    table_definitions[ref] = table_definitions[ref].expand_columns(
                        table_definitions=table_definitions,
                        referenced=True)
                    col = FKColumn(
                        table_ref=table_definitions[ref],
                        name=col_name,
                        database_flavor=self.database_flavor,
                    )
                elif col_object['$ref'] in columns_definitions:
                    logging.debug(
                        'Column ref a type that is not a object. '
                        'Copy Column from columns definitions')
                    ref = col_object['$ref']
                    ref_col = columns_definitions[ref]
                    col_as_dict = {**asdict(ref_col), 'name': col_name}
                    col = Column(**col_as_dict)
                else:
                    logging.debug('Skipping ref as it is not in table definitions neither in columns definitions')
                    continue
            else:
                col = Column(
                    name=col_name,
                    database_flavor=self.database_flavor,
                    jsonschema_type=col_object['type'],
                    jsonschema_fields=col_object,
                )
            self.columns.add(col)
            if col.is_pk:
                self.primary_key = col

            self.logger.info(f'New created column {col}')

        if referenced and not self.primary_key:
            self.logger.info('Creating id column for the table in order to reference it as PK')
            col = Column(
                name='id',
                database_flavor=self.database_flavor,
                jsonschema_type='id',
            )
            self.columns.add(col)
            self.primary_key = col

        return self


@dataclass(eq=False)
class FKColumn(Column):
    table_ref: Table = None

    @property
    def data_type(self) -> str:
        data_type_ref = self.table_ref.primary_key.data_type
        if "varchar" in data_type_ref:
            return data_type_ref
        else:
            return FK_TYPES.get(data_type_ref, 'bigint')

    @staticmethod
    def is_fk() -> bool:
        """Returns true if the column is a foreign key.

        Returns:
            bool: True if it is foreign key
        """
        return True

    # FIXME: Avoid overwritting the the repr method
    # NOTE: Overwrite dataclass method to show data_type property
    def __repr__(self):
        return f"FKColumn(name={self.name} data_type={self.data_type} table_ref.name={self.table_ref.name})"



import logging
import os
from typing import Dict, List

from jsonschema2ddl.models import Column, Table
from jsonschema2ddl.utils import db_column_name, db_table_name


class JSONSchemaToDatabase:
    """JSONSchemaToDatabase is the mother class for everything.

    Typically you want to instantiate a `JSONSchemaToPostgres` object, and
    run :func:`create_tables` to create all the tables. Run :func:`create_links`
    to populate all references properly and add foreign keys between tables.
    Optionally you can run :func:`analyze` finally which optimizes the tables.
    """

    logger: logging.Logger = logging.getLogger('JSONSchemaToDatabase')

    def __init__(
            self,
            schema: Dict,
            database_flavor: str = "postgres",
            schema_name: str = None,  # postgres_schema
            abbreviations: Dict = {},  # TODO: Implement abbreviations
            extra_columns: List = [],  # TODO: Implement extra columns
            root_table_name: str = 'root',
            log_level: str = os.getenv('LOG_LEVEL', 'DEBUG')):

        self.logger.setLevel(log_level)
        # Table.logger = self.logger.getChild('Table')
        # Column.logger = self.logger.getChild('Column')

        self.schema = schema

        self.database_flavor = database_flavor
        self.schema_name = schema_name
        self.root_table_name = db_table_name(root_table_name, schema_name=self.schema_name)
        self.extra_columns = extra_columns
        self.abbreviations = abbreviations

        self.table_definitions = self._create_table_definitions()
        self.logger.info('Table definitions initialized')

    def _create_table_definitions(self):

        # NOTE: create first empty tables to reference later in columns
        table_definitions = dict()
        columns_definitions = dict()
        schema_definitions = self.schema.get('definitions', {})
        for name, object_schema in schema_definitions.items():
            ref = object_schema.get("$id") or f"#/definitions/{name}"
            if object_schema['type'] == 'object':
                table = Table(
                    ref=ref,
                    name=db_table_name(name, schema_name=self.schema_name),
                    comment=object_schema.get('comment'),
                    jsonschema_fields=object_schema,
                )
                table_definitions[table.ref] = table
            else:
                # NOTE: Create new column for main table
                column = Column(
                    name=db_column_name(name),
                    database_flavor=self.database_flavor,
                    jsonschema_type=object_schema['type'],
                    jsonschema_fields=object_schema,
                )
                columns_definitions[ref] = column

        root_table = Table(
            ref='root',
            name=self.root_table_name,
            comment=self.schema.get('comment', ""),
            jsonschema_fields=self.schema,
        )
        table_definitions[root_table.ref] = root_table

        for ref, table in table_definitions.items():
            table_definitions[ref] = table.expand_columns(table_definitions, columns_definitions)

        return table_definitions

    def _execute(self, cursor, query, args=None, query_ok_to_print=True):
        self.logger.debug(query)
        cursor.execute(query, args)

    def create_tables(
            self,
            conn,
            drop_schema: bool = False,
            drop_tables: bool = False,
            drop_cascade: bool = True,
            auto_commit: bool = False):

        with conn.cursor() as cursor:
            self.logger.info(f'Creating tables in the schema {self.schema_name}')
            if self.schema_name is not None:
                if drop_schema:
                    self.logger.info(f'Dropping schema {self.schema_name}!!')
                    self._execute(
                        cursor,
                        f'DROP SCHEMA IF EXISTS {self.schema_name} {"CASCADE;" if drop_cascade else ";"}'
                    )
                self._execute(cursor, f'CREATE SCHEMA IF NOT EXISTS {self.schema_name};')

            self.logger.debug(self.table_definitions.keys())
            for table_ref, table in self.table_definitions.items():
                # FIXME: Move to a separate method
                self.logger.info(f'Trying to create table {table.name}')
                self.logger.debug(table_ref)
                self.logger.debug(table)
                if drop_tables:
                    self.logger.info(f'Dropping table {table.name}!!')
                    self._execute(
                        cursor,
                        f'DROP TABLE IF EXISTS {table.name} {"CASCADE;" if drop_cascade else ";"}'
                    )
                all_cols = [f' "{col.name}" {col.data_type}' for col in table.columns]
                unique_cols = [f'"{col}"' for col in table.columns if col.is_unique]
                create_q = (
                    f"""CREATE TABLE {table.name} ( """
                    f"""{','.join(all_cols)} """
                    f"""{", UNIQUE (" + ','.join(unique_cols) +  ")" if len(unique_cols) > 0 else ""} """
                    f"""{", PRIMARY KEY (" + table.primary_key.name +  ")" if table.primary_key else ""}); """
                )
                self._execute(cursor, create_q)
                if table.comment:
                    self.logger.debug(f'Set the following comment on table {table.name}: {table.comment}')
                    self._execute(cursor, f"COMMENT ON TABLE {table.name} IS '{table.comment}'")
                for col in table.columns:
                    if col.comment:
                        self.logger.debug(f'Set the following comment on column {col.name}: {col.comment}')
                        self._execute(cursor, f'COMMENT ON COLUMN {table.name}."{col.name}" IS ' + f"'{col.comment}'")
                self.logger.info('Table created!')

        if auto_commit:
            conn.commit()

    def create_links(self, conn, auto_commit: bool = True):
        """Adds foreign keys between tables.

        Args:
            conn (): connection object
            auto_commit (bool, Optional): Defaults to False
        """
        for table_ref, table in self.table_definitions.items():
            print(table_ref)
            print(table)
            for col in table.columns:
                print(col)
                if col.is_fk():
                    fk_q = (
                        f"""ALTER TABLE {table.name} """
                        f"""ADD CONSTRAINT fk_{col.table_ref.name.split('"')[-2]} """  # FIXME: Formatting hack
                        f"""FOREIGN KEY ({col.name}) """
                        f"""REFERENCES {col.table_ref.name} ({col.table_ref.primary_key.name}); """
                    )
                    with conn.cursor() as cursor:
                        self._execute(cursor, fk_q)
                    if auto_commit:
                        conn.commit()
        # raise Exception('OH NO')

    def analyze(self, conn):
        """Runs `analyze` on each table. This improves performance.

        See the `Postgres documentation for Analyze
        <https://www.postgresql.org/docs/9.1/static/sql-analyze.html>`_

        Args:
            conn (): connection object
        """
        self.logger.info('Analyzing tables...')
        with conn.cursor() as cursor:
            for table_ref, table in self.table_definitions.items():
                self.logger.info(f'Launch analyze for {table.name}')
                self._execute(cursor, 'ANALYZE %s' % table.name)


class JSONSchemaToPostgres(JSONSchemaToDatabase):
    """Shorthand for JSONSchemaToDatabase(..., database_flavor='postgres')"""

    def __init__(self, *args, **kwargs):
        kwargs['database_flavor'] = 'postgres'
        return super(JSONSchemaToPostgres, self).__init__(*args, **kwargs)


class JSONSchemaToRedshift(JSONSchemaToDatabase):
    """Shorthand for JSONSchemaToDatabase(..., database_flavor='redshift')"""

    def __init__(self, *args, **kwargs):
        kwargs['database_flavor'] = 'redshift'
        return super(JSONSchemaToRedshift, self).__init__(*args, **kwargs)
